"""Specific views for keywords / classification schemes

:organization: Logilab
:copyright: 2003-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from logilab.mtconverter import xml_escape

from cubicweb import Unauthorized
from cubicweb.predicates import is_instance, rql_condition, relation_possible
from cubicweb.view import EntityView
from cubicweb.web import stdmsgs, component, facet
from cubicweb.web.views import primary, treeview
from cubicweb.web.views.ajaxcontroller import ajaxfunc

from cubicweb.web.views import uicfg

_pvs = uicfg.primaryview_section
_pvs.tag_object_of(('*', 'applied_to', '*'), 'hidden')
_pvs.tag_subject_of(('*', 'applied_to', '*'), 'hidden')
_pvs.tag_object_of(('*', 'included_in', 'Classification'), 'hidden')

_abaa = uicfg.actionbox_appearsin_addmenu
_abaa.tag_object_of(('CodeKeyword', 'included_in', 'Classification'), True)
_abaa.tag_object_of(('Keyword', 'included_in', 'Classification'), True)
_abaa.tag_object_of(('CodeKeyword', 'subkeyword_of', 'CodeKeyword'), True)
_abaa.tag_object_of(('Keyword', 'subkeyword_of', 'Keyword'), True)


# classification views ########################################################

class ClassificationPrimaryView(primary.PrimaryView):
    __select__ = is_instance('Classification')

    def render_entity_attributes(self, entity):
        pass

    def render_entity_relations(self, entity):
        rset = self._cw.execute('Any K ORDERBY N WHERE K included_in C, '
                                'NOT K subkeyword_of KK, K name N, '
                                'C eid %(x)s', {'x': entity.eid})
        self.wview('treeview', rset, 'null')


# keyword views ###############################################################

class KeywordPrimaryView(primary.PrimaryView):
    __select__ = is_instance('Keyword')

    def cell_call(self, row, col, **kwargs):
        entity = self.cw_rset.get_entity(row, col)
        self.w(u'<h1 class="titleUnderline">%s</h1>'
               % xml_escape(entity.dc_long_title()))
        rset = entity.related('subkeyword_of', 'object')
        self.wview('treeview', rset, 'null')


class KeywordComboBoxView(treeview.TreePathView):
    """display keyword in edition's combobox"""
    __regid__ = 'combobox'
    __select__ = is_instance('Keyword', 'Classification')

    item_vid = 'text'
    separator = u' > '

# skos views ############################################################


SKOS_OPENING_ROOT = u'''<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE rdfs [
    <!ENTITY rdf "http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    <!ENTITY rdfs "http://www.w3.org/2000/01/rdf-schema#">
    <!ENTITY dc "http://purl.org/dc/elements/1.1/">
    <!ENTITY dct "http://purl.org/dc/terms/">
]>
<rdf:RDF  xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
          xmlns:skos="http://www.w3.org/2004/02/skos/core#">
'''
SKOS_CLOSING_ROOT = u'</rdf:RDF>'


class SkosView(EntityView):
    __regid__ = 'skos'
    content_type = 'application/xml'
    templatable = False
    __select__ = is_instance('Keyword', 'Classification')

    def call(self, **kwargs):
        self.w(SKOS_OPENING_ROOT)
        for i in range(self.rset.rowcount):
            self.cell_call(i, 0)
        self.w(SKOS_CLOSING_ROOT)

    def cell_call(self, row, col):
        self.wview('skositemview', self.rset, row=row, col=col)


class SkosItemView(EntityView):
    __regid__ = 'skositemview'
    content_type = 'application/xml'
    __select__ = is_instance('Keyword', 'Classification')

    def cell_call(self, row, col, show_parent=True, stop=False):
        w = self.w
        entity = self.complete_entity(row, col)
        eschema = entity.e_schema
        w(u'<skos:%s>' % eschema)
        w(u'<skos:prefLabel>%s</skos:prefLabel>' % xml_escape(entity.name))
        if not stop:
            if show_parent and not entity.is_root():
                par = entity.parent()
                w(u'<skos:broader>')
                par.view('skositemview', show_parent=False, stop=True, w=self.w)
                w(u'</skos:broader>')
            for child in entity.children(entities=True):
                w(u'<skos:narrower>')
                self.wview('skositemview', child.as_rset(), show_parent=False)
                w(u'</skos:narrower>')
        w(u'</skos:%s>' % eschema)


# keyword component ###########################################################

class KeywordBarVComponent(component.EntityCtxComponent):
    """the keywords path bar: display keywords of a tagged entity"""
    __regid__ = 'keywordsbar'
    __select__ = (component.EntityCtxComponent.__select__ &
                  relation_possible('applied_to', 'object', 'Keyword'))
    context = 'header'
    order = 152
    htmlclass = 'navigation'

    def get_keywords(self):
        """helper method for subclasses redefinition"""
        return self.entity.related('applied_to', 'object')

    def render_body(self, w):
        rset = self.get_keywords()
        if rset:
            w(u'<div class="%s" id="%s">\n' % (self.cssclass, self.domid))
            w(u'<span>%s</span>&nbsp;' % self._cw._('keywords:'))
            self._cw.view('csv', rset, 'null', w=w)
            w(u'</div>\n')
        else:
            w(u'<div class="%s hidden" id="%s"></div>\n' % (
                self.cssclass, self.domid))


class AddKeywordVComponent(component.EntityCtxComponent):
    """the 'add keyword' component"""
    __regid__ = 'addkeywords'
    __select__ = component.EntityCtxComponent.__select__ & \
        relation_possible('applied_to', 'object', 'Keyword', action='add') & \
        rql_condition('X is ET, CL classifies ET')

    context = 'header'
    order = 153
    htmlclass = 'navigation'

    def render_body(self, w):
        entity = self.entity
        self._cw.add_js(['cubicweb.widgets.js', 'cubes.keyword.js'])
        self._cw.add_css('cubicweb.suggest.css')
        w(u'<table><tr><td>')
        w(u'<a class="button sglink" href="javascript: showKeywordSelector(%s, \'%s\', \'%s\');">%s</a></td>' % (
            entity.eid, self._cw._(stdmsgs.BUTTON_OK[0]),
            self._cw._(stdmsgs.BUTTON_CANCEL[0]), self._cw._('add keywords')))
        w(u'<td><div id="kwformholder"></div>')
        w(u'</td></tr></table>')


# applied_to relation facet ####################################################

class AppliedToFacet(facet.RelationFacet):
    __regid__ = 'applied-to-facet'
    rtype = 'applied_to'
    role = 'object'
    target_attr = 'name'

    def rset_vocabulary(self, rset):
        _ = self._cw._
        vocab = []
        scheme = None
        for e in sorted(rset.entities(),
                        key=lambda e: (e.cw_adapt_to('ITree').classification.name,
                                       e.view('combobox'))):
            classification_name = e.cw_adapt_to('ITree').classification.name
            if scheme != classification_name:
                vocab.append((_(classification_name), None))
            vocab.append((e.view('combobox'), e.eid))
        return vocab


class ClassificationFacet(facet.RelationFacet):
    """abstract per-classification facet

    subclasses must define their own id the classification name, e.g :

    class Classifaction1Facet(ClassificationFacet):
        __regid__ = 'classif1'
        classification = u'classification1'

    """
    __abstract__ = True
    classification = None
    rtype = 'applied_to'
    role = 'object'

    def vocabulary(self):
        """return vocabulary for this facet, eg a list of 2-uple (label, value)
        """
        rqlst = self.rqlst
        rqlst.save_state()
        try:
            mainvar = self.filtered_variable  # X
            keyword_var = rqlst.make_variable()  # K
            keyword_name_var = rqlst.make_variable()  # KN
            classif_var = rqlst.make_variable()  # C
            rqlst.make_variable()  # CN
            rqlst.add_relation(keyword_var, 'applied_to', mainvar)  # K applied_to X
            rqlst.add_relation(keyword_var, 'name', keyword_name_var)  # K name KN
            rqlst.add_relation(keyword_var, 'included_in', classif_var)  # K included_in C
            # C name "classification-name"
            rqlst.add_constant_restriction(classif_var, 'name', self.classification, 'String')
            rqlst.add_selected(keyword_var)
            rqlst.add_selected(keyword_name_var)
            # ORDERBY KN
            rqlst.add_sort_var(keyword_name_var, True)
            try:
                rset = self._cw.execute(rqlst.as_string(), self.rset.args,
                                        self.rset.cachekey)
            except Unauthorized:
                return []
        finally:
            rqlst.recover()
        return self.rset_vocabulary(rset)

    @property
    def title(self):
        return self._cw._(self.classification)

    def support_and(self):
        return False

# add some classification schema related methods to the Jsoncontroller ########


@ajaxfunc(output_type='json')
def js_possible_keywords(self, eid):
    rql = ('DISTINCT Any N WHERE K is Keyword, K name N, NOT K applied_to X, '
           'X eid %(x)s, K included_in C, C classifies ET, X is ET')
    rset = self.cursor.execute(rql, {'x': eid, 'u': self._cw.user.eid}, 'x')
    return [name for (name,) in rset]


@ajaxfunc(output_type='json')
def js_add_keywords(self, eid, kwlist):
    msg = self._cw._('keywords applied')
    kwrset = self.cursor.execute('Any K,N,C WHERE K is Keyword, K name N, K included_in C, '
                                 'C classifies ET, X eid %(x)s, X is ET',
                                 {'x': eid}, 'x')
    if not kwrset:
        return self._cw._('No suitable classification scheme found')
    classification = kwrset[0][2]  # XXX what if we have several classifications ?
    valid_keywords = set(kwname for _, kwname, _ in kwrset)
    user_keywords = set(kwlist)
    invalid_keywords = sorted(user_keywords - valid_keywords)
    kweids = dict((kwname, str(kweid)) for kweid, kwname, _ in kwrset if kwname in user_keywords)
    if invalid_keywords:
        for keyword in invalid_keywords:
            neweid = self.cursor.execute('INSERT Keyword K: K name %(name)s, K included_in C WHERE C eid %(c)s',
                                         {'name': keyword, 'c': classification}, 'c')[0][0]
            kweids[keyword] = str(neweid)
    if kweids:
        self.cursor.execute('SET KW applied_to X WHERE X eid %%(x)s, KW eid IN (%s)'
                            % ','.join(kweids.values()), {'x': eid}, 'x')
    return msg



"""specific hooks for Classification and Keyword entities

:organization: Logilab
:copyright: 2007-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

from itertools import chain

from cubicweb import ValidationError
from cubicweb.server.hook import Hook, Operation, match_rtype
from yams.schema import role_name
from cubicweb import _


class BeforeAddDescendantOf(Hook):
    """check indirect cycle for ``descendant_of`` relation
    """
    __regid__ = 'beforeadddescendant'
    events = ('before_add_relation', )
    __select__ = Hook.__select__ & match_rtype('descendant_of')

    def __call__(self):
        entity = self._cw.entity_from_eid(self.eidfrom)
        parent = self._cw.entity_from_eid(self.eidto)
        parents = set([x.eid for x in chain([parent, ], parent.cw_adapt_to('ITree').iterparents())])
        children = set([x.eid for x in chain([entity], entity.cw_adapt_to('ITree').recurse_children())])
        if children & parents:
            msg = _('detected descendant_of cycle')
            raise ValidationError(self.eidfrom, {role_name(self.rtype, 'subject'): msg})


class AfterAddSubKeywordOf(Hook):
    """sets ``descendant_of`` relation
    """
    __regid__ = 'afteradddescendant'
    events = ('after_add_relation', )
    __select__ = Hook.__select__ & match_rtype('subkeyword_of')

    def __call__(self):
        entity = self._cw.entity_from_eid(self.eidfrom)
        parent = self._cw.entity_from_eid(self.eidto)
        SetDescendantOfKeywordOp(self._cw, parent=parent, entity=entity)


class SetIncludedInRelationHook(Hook):
    """sets the included_in relation on a subkeyword if not already set
    """
    __regid__ = 'setincludedinrelationhook'
    events = ('before_add_relation',)
    __select__ = Hook.__select__ & match_rtype('subkeyword_of')

    def __call__(self):
        # immediate test direct cycles
        if self.eidfrom == self.eidto:
            msg = self._cw._('keyword cannot be subkeyword of himself')
            raise ValidationError(self.eidfrom, {role_name(self.rtype, 'subject'): msg})
        SetIncludedInRelationOp(self._cw, vreg=self._cw.vreg,
                                eidfrom=self.eidfrom, eidto=self.eidto)


class RemoveDescendantOfRelation(Hook):
    """removes ``descendant_of`` relation

    we delete the relation for entity's parents recursively
    """
    __regid__ = 'removedescendantofrelation'
    events = ('after_delete_relation',)
    __select__ = Hook.__select__ & match_rtype('subkeyword_of')

    def __call__(self):
        parent = self._cw.entity_from_eid(self.eidto)
        for parent in chain([parent], parent.cw_adapt_to('ITree').iterparents()):
            self._cw.execute('DELETE K descendant_of P WHERE K eid %(k)s, '
                             'P eid %(p)s', {'p': parent.eid, 'k': self.eidfrom})


# operations #################################################################
class SetIncludedInRelationOp(Operation):
    """delay this operation to commit to avoid conflict with a late rql query
    already setting the relation
    """

    def precommit_event(self):
        cnx = self.cnx
        # test for indirect cycles
        self.check_cycle()
        subkw = cnx.entity_from_eid(self.eidfrom)
        if subkw.included_in:
            kw = cnx.entity_from_eid(self.eidto)
            if subkw.included_in[0].eid != kw.included_in[0].eid:
                msgid = "keywords %(subkw)s and %(kw)s belong to different classifications"
                raise ValidationError(subkw.eid, {role_name('subkeyword_of', 'subject'): cnx._(msgid) %
                                                  {'subkw': subkw.eid, 'kw': kw.eid}})
        else:
            cnx.execute('SET SK included_in C WHERE SK eid %(x)s, '
                        'SK subkeyword_of K, K included_in C',
                        {'x': subkw.eid})

    def check_cycle(self):
        parents = set([self.eidto])
        parent = self.cnx.entity_from_eid(self.eidto)
        while parent.subkeyword_of:
            parent = parent.subkeyword_of[0]
            if parent.eid in parents:
                msg = self.cnx._('detected subkeyword cycle')
                raise ValidationError(self.eidfrom, {role_name('subkeyword_of', 'subject'): msg})
            parents.add(parent.eid)


class SetDescendantOfKeywordOp(Operation):
    def precommit_event(self):
        """transitive closure of ``descendant_of`` relations to current entity"""
        closure = set()
        entity = self.entity
        parent = self.parent
        for parent in chain([parent, entity], parent.cw_adapt_to('ITree').iterparents()):
            for child in chain([entity], entity.cw_adapt_to('ITree').recurse_children()):
                if child.eid != parent.eid:
                    closure.add((child, parent))
        for child, parent in closure:
            descendants = [p.eid for p in child.descendant_of]
            if parent.eid not in descendants:
                child.cw_set(descendant_of=parent)



from cubicweb.devtools.testlib import CubicWebTC


class SecurityTC(CubicWebTC):
    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            cwgroup = cnx.execute('Any U WHERE U is ET, U name "CWGroup"').one()
            self.classif1_eid = cnx.create_entity('Classification', name=u"classif1", classifies=cwgroup).eid
            self.kw1_eid = cnx.create_entity('Keyword', name=u'kw1', included_in=self.classif1_eid).eid
            cnx.commit()

    def test_nonregr_keyword_selection_as_guest(self):
        with self.new_access("anon").cnx() as cnx:
            cnx.execute('Any X ORDERBY Z WHERE X modification_date Z, K eid %(k)s, K applied_to X', {'k': self.kw1_eid})


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



from cubicweb.devtools.testlib import CubicWebTC

from cubicweb import ValidationError


class ClassificationHooksTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            # classification for CWGroup
            cwgroup = cnx.execute('Any U WHERE U is ET, U name "CWGroup"').one()
            self.classif1_eid = cnx.create_entity('Classification', name=u"classif1", classifies=cwgroup).eid
            self.kwgroup_eid = cnx.create_entity('Keyword', name=u'kwgroup', included_in=self.classif1_eid).eid
            # classification for CWUser
            cwuser = cnx.execute('Any U WHERE U is ET, U name "CWUser"').one()
            self.classif2_eid = cnx.create_entity('Classification', name=u"classif2", classifies=cwuser).eid
            self.kwuser_eid = cnx.create_entity('Keyword', name=u'kwuser', included_in=self.classif2_eid).eid
            cnx.commit()

    def test_application_of_bad_keyword_fails(self):
        with self.admin_access.cnx() as cnx:
            cnx.execute('SET K applied_to G WHERE G is CWGroup, K eid %(k)s',
                        {'k': self.kwuser_eid})
            self.assertRaises(ValidationError, cnx.commit)

    def test_creating_a_new_subkeyword_sets_included_in(self):
        with self.admin_access.cnx() as cnx:
            kwgroup2_eid = cnx.create_entity('Keyword', name=u'kwgroup2', subkeyword_of=self.kwgroup_eid).eid
            cnx.commit()
            rset = cnx.execute('Any N WHERE C name N, K included_in C, K eid %(k)s', {'k': kwgroup2_eid})
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset[0][0], 'classif1')

    def test_cannot_create_subkeyword_from_other_classification(self):
        with self.admin_access.cnx() as cnx:
            cnx.execute('SET K1 subkeyword_of K2 WHERE K1 eid %(k1)s, K2 eid %(k2)s',
                        {'k1': self.kwgroup_eid, 'k2': self.kwuser_eid})
            self.assertRaises(ValidationError, cnx.commit)

    def test_cannot_create_cycles(self):
        with self.admin_access.cnx() as cnx:
            # direct obvious cycle
            cnx.execute('SET K1 subkeyword_of K2 WHERE K1 eid %(k1)s, K2 eid %(k2)s',
                        {'k1': self.kwgroup_eid, 'k2': self.kwuser_eid})
            self.assertRaises(ValidationError, cnx.commit)
            # testing indirect cycles
            kwgroup2_eid = cnx.create_entity('Keyword', name=u'kwgroup2',
                                             included_in=self.classif1_eid,
                                             subkeyword_of=self.kwgroup_eid).eid
            cnx.execute('SET K subkeyword_of K2 WHERE K eid %(k)s, K2 eid %(k2)s',
                        {'k': self.kwgroup_eid, 'k2': kwgroup2_eid})
            self.assertRaises(ValidationError, cnx.commit)


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



from __future__ import with_statement
from cubicweb.devtools.testlib import CubicWebTC

from cubicweb import ValidationError


class KeywordHooksTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            cwgroup = cnx.execute('Any U WHERE U is ET, U name "CWGroup"').one()
            self.classif1_eid = cnx.create_entity('Classification', name=u"classif1", classifies=cwgroup).eid
            cnx.commit()

    def test_keyword_add1(self):
        with self.admin_access.cnx() as cnx:
            ce = cnx.create_entity
            kw1_eid = ce('Keyword', name=u'kw1', included_in=self.classif1_eid).eid
            kw2_eid = ce('Keyword', name=u'kw2', subkeyword_of=kw1_eid, included_in=self.classif1_eid).eid
            kw3_eid = ce('Keyword', name=u'kw3', subkeyword_of=kw2_eid, included_in=self.classif1_eid).eid
            kw4_eid = ce('Keyword', name=u'kw4', subkeyword_of=kw3_eid, included_in=self.classif1_eid).eid
            kw5_eid = ce('Keyword', name=u'kw5', subkeyword_of=kw4_eid, included_in=self.classif1_eid).eid
            cnx.commit()
            parent = cnx.find('Keyword', eid=kw1_eid).one()
            child = cnx.find('Keyword', eid=kw5_eid).one()
            self.assertCountEqual([kw.name for kw in child.cw_adapt_to('ITree').iterparents()],
                                  ['kw4', 'kw3', 'kw2', 'kw1'])
            self.assertCountEqual([kw.name for kw in child.descendant_of], ['kw4', 'kw3', 'kw2', 'kw1'])
            self.assertCountEqual([kw.name for kw in parent.reverse_descendant_of], ['kw5', 'kw4', 'kw3', 'kw2'])
            self.assertCountEqual([kw.name for kw in parent.cw_adapt_to('ITree').recurse_children()],
                                  ['kw5', 'kw4', 'kw3', 'kw2'])

    def test_keyword_add2(self):
        with self.admin_access.cnx() as cnx:
            ce = cnx.create_entity
            kw1_eid = ce('Keyword', name=u'kw1', included_in=self.classif1_eid).eid
            kw2_eid = ce('Keyword', name=u'kw2', subkeyword_of=kw1_eid, included_in=self.classif1_eid).eid
            kw3_eid = ce('Keyword', name=u'kw3', included_in=self.classif1_eid).eid
            kw4_eid = ce('Keyword', name=u'kw4', subkeyword_of=kw3_eid, included_in=self.classif1_eid).eid
            kw5_eid = ce('Keyword', name=u'kw5',  subkeyword_of=kw4_eid, included_in=self.classif1_eid).eid
            cnx.commit()
            cnx.execute('SET K3 subkeyword_of K2 WHERE K3 eid %(kw3)s, K2 eid %(kw2)s',
                        {'kw3': kw3_eid, 'kw2': kw2_eid})
            cnx.commit()
            parent = cnx.find('Keyword', eid=kw1_eid).one()
            child = cnx.find('Keyword', eid=kw5_eid).one()
            self.assertCountEqual([kw.name for kw in child.cw_adapt_to('ITree').iterparents()],
                                  ['kw4', 'kw3', 'kw2', 'kw1'])
            self.assertCountEqual([kw.name for kw in child.descendant_of], ['kw4', 'kw3', 'kw2', 'kw1'])
            self.assertCountEqual([kw.name for kw in parent.reverse_descendant_of], ['kw5', 'kw4', 'kw3', 'kw2'])
            self.assertCountEqual([kw.name for kw in parent.cw_adapt_to('ITree').recurse_children()],
                                  ['kw5', 'kw4', 'kw3', 'kw2'])

    def test_keyword_add3(self):
        with self.admin_access.cnx() as cnx:
            ce = cnx.create_entity
            kw1_eid = ce('Keyword', name=u'kw1', included_in=self.classif1_eid).eid
            kw2_eid = ce('Keyword', name=u'kw2', subkeyword_of=kw1_eid, included_in=self.classif1_eid).eid
            kw3_eid = ce('Keyword', name=u'kw3', included_in=self.classif1_eid).eid
            kw4_eid = ce('Keyword', name=u'kw4', included_in=self.classif1_eid).eid
            kw5_eid = ce('Keyword', name=u'kw5', subkeyword_of=kw4_eid, included_in=self.classif1_eid).eid
            cnx.commit()
            cnx.execute('SET K2 subkeyword_of K3 WHERE K2 eid %(k2)s, K3 eid %(k3)s',
                        {'k2': kw4_eid, 'k3': kw3_eid})
            cnx.execute('SET K3 subkeyword_of K4 WHERE K3 eid %(k3)s, K4 eid %(k4)s',
                        {'k3': kw3_eid, 'k4': kw2_eid})
            cnx.commit()
            child = cnx.find('Keyword', eid=kw5_eid).one()
            parent = cnx.find('Keyword', eid=kw1_eid).one()
            self.assertCountEqual([kw.name for kw in child.descendant_of], ['kw4', 'kw3', 'kw2', 'kw1'])
            # XXX check the order of iterparents
            self.assertCountEqual([kw.name for kw in child.cw_adapt_to('ITree').iterparents()],
                                  ['kw4', 'kw3', 'kw2', 'kw1'])
            self.assertCountEqual([kw.name for kw in parent.cw_adapt_to('ITree').recurse_children()],
                                  ['kw2', 'kw3', 'kw4', 'kw5'])
            self.assertCountEqual([kw.name for kw in parent.reverse_descendant_of], ['kw2', 'kw3', 'kw4', 'kw5'])

    def test_keyword_add4(self):
        with self.admin_access.cnx() as cnx:
            ce = cnx.create_entity
            kw0_eid = ce('Keyword', name=u'kw0', included_in=self.classif1_eid).eid
            kw1_eid = ce('Keyword', name=u'kw1', included_in=self.classif1_eid).eid
            kw2_eid = ce('Keyword', name=u'kw2', subkeyword_of=kw1_eid, included_in=self.classif1_eid).eid
            kw3_eid = ce('Keyword', name=u'kw3', included_in=self.classif1_eid).eid
            kw4_eid = ce('Keyword', name=u'kw4', included_in=self.classif1_eid).eid
            kw5_eid = ce('Keyword', name=u'kw5',  subkeyword_of=kw4_eid, included_in=self.classif1_eid).eid
            cnx.execute('SET K3 subkeyword_of K2 WHERE K3 eid %(kw3)s, K2 eid %(kw2)s',
                        {'kw2': kw2_eid, 'kw3': kw3_eid})
            cnx.commit()
            kw3 = cnx.find('Keyword', eid=kw3_eid).one()
            self.assertCountEqual([kw.name for kw in kw3.descendant_of], ['kw1', 'kw2'])
            cnx.execute('SET K3 descendant_of K0 WHERE K3 eid %(kw3)s, K0 eid %(kw0)s',
                        {'kw3': kw3_eid, 'kw0': kw0_eid})
            cnx.commit()
            kw3 = cnx.find('Keyword', eid=kw3_eid).one()
            self.assertCountEqual([kw.name for kw in kw3.descendant_of], ['kw0', 'kw1', 'kw2'])
            cnx.execute('SET K3 descendant_of K4 WHERE K3 eid %(kw3)s, K4 eid %(kw4)s',
                        {'kw3': kw3_eid, 'kw4': kw4_eid})
            cnx.commit()
            kw3 = cnx.find('Keyword', eid=kw3_eid).one()
            self.assertCountEqual([kw.name for kw in kw3.descendant_of], ['kw0', 'kw1', 'kw2', 'kw4'])
            cnx.execute('SET K3 descendant_of K5 WHERE K3 eid %(kw3)s, K5 eid %(kw5)s',
                        {'kw3': kw3_eid, 'kw5': kw5_eid})
            cnx.commit()
            kw3.cw_clear_all_caches()
            self.assertCountEqual([kw.name for kw in kw3.descendant_of], ['kw0', 'kw1', 'kw2', 'kw4', 'kw5'])

    def test_keyword_update1(self):
        with self.admin_access.cnx() as cnx:
            ce = cnx.create_entity
            kw1_eid = ce('Keyword', name=u'kw1', included_in=self.classif1_eid).eid
            kw2_eid = ce('Keyword', name=u'kw2', subkeyword_of=kw1_eid, included_in=self.classif1_eid).eid
            kw3_eid = ce('Keyword', name=u'kw3', included_in=self.classif1_eid).eid
            kw4_eid = ce('Keyword', name=u'kw4', included_in=self.classif1_eid).eid
            kw5_eid = ce('Keyword', name=u'kw5',  subkeyword_of=kw4_eid, included_in=self.classif1_eid).eid
            cnx.execute('SET K3 subkeyword_of K2 WHERE K3 eid %(kw3)s, K2 eid %(kw2)s',
                        {'kw3': kw3_eid, 'kw2': kw2_eid})
            cnx.commit()
            kw3 = cnx.find('Keyword', eid=kw3_eid).one()
            self.assertCountEqual([kw.name for kw in kw3.descendant_of], ['kw1', 'kw2'])
            cnx.execute('SET K3 subkeyword_of K4 WHERE K3 eid %(kw3)s, K4 eid %(kw4)s',
                        {'kw3': kw3_eid, 'kw4': kw4_eid})
            cnx.commit()
            kw3 = cnx.find('Keyword', eid=kw3_eid).one()
            self.assertCountEqual([kw.name for kw in kw3.descendant_of], ['kw4'])
            cnx.execute('SET K3 subkeyword_of K5 WHERE K3 eid %(kw3)s, K5 eid %(kw5)s',
                        {'kw3': kw3_eid, 'kw5': kw5_eid})
            cnx.commit()
            kw3 = cnx.find('Keyword', eid=kw3_eid).one()
            self.assertCountEqual([kw.name for kw in kw3.descendant_of], ['kw4', 'kw5'])

    def test_keyword_descendant_of(self):
        with self.admin_access.cnx() as cnx:
            ce = cnx.create_entity
            kw1_eid = ce('Keyword', name=u'kw1', included_in=self.classif1_eid).eid
            kw2_eid = ce('Keyword', name=u'kw2', subkeyword_of=kw1_eid, included_in=self.classif1_eid).eid
            kw3_eid = ce('Keyword', name=u'kw3', subkeyword_of=kw1_eid, included_in=self.classif1_eid).eid
            cnx.commit()

            kw1 = cnx.find('Keyword', eid=kw1_eid).one()
            kw2 = cnx.find('Keyword', eid=kw2_eid).one()
            kw3 = cnx.find('Keyword', eid=kw3_eid).one()
            self.assertCountEqual([kw.name for kw in kw2.descendant_of], ['kw1', ])
            self.assertCountEqual([kw.name for kw in kw3.descendant_of], ['kw1', ])
            self.assertCountEqual([kw.name for kw in kw1.reverse_descendant_of], ['kw3', 'kw2'])
            self.assertCountEqual([kw.name for kw in kw1.cw_adapt_to('ITree').recurse_children()], ['kw2', 'kw3'])
            kw0_eid = ce('Keyword', name=u'kw0', included_in=self.classif1_eid).eid

            cnx.execute('SET K1 subkeyword_of K0 WHERE K1 eid %(kw1)s, K0 eid %(kw0)s',
                        {'kw1': kw1_eid, 'kw0': kw0_eid})
            cnx.commit()

            kw0 = cnx.find('Keyword', eid=kw0_eid).one()
            kw1 = cnx.find('Keyword', eid=kw1_eid).one()
            kw2 = cnx.find('Keyword', eid=kw2_eid).one()
            kw3 = cnx.find('Keyword', eid=kw3_eid).one()
            self.assertCountEqual([kw.name for kw in kw0.cw_adapt_to('ITree').recurse_children()],
                                  ['kw1', 'kw2', 'kw3'])
            self.assertCountEqual([kw.name for kw in kw0.reverse_descendant_of], ['kw3', 'kw2', 'kw1'])
            self.assertCountEqual([kw.name for kw in kw1.descendant_of], ['kw0'])
            self.assertCountEqual([kw.name for kw in kw2.descendant_of], ['kw1', 'kw0'])
            self.assertCountEqual([kw.name for kw in kw3.descendant_of], ['kw1', 'kw0'])

    def test_keyword_delete(self):
        """*after_delete_relation* of ``subkeyword_of``
        """
        with self.admin_access.cnx() as cnx:
            ce = cnx.create_entity
            kw1_eid = ce('Keyword', name=u'kw1', included_in=self.classif1_eid).eid
            kw2_eid = ce('Keyword', name=u'kw2', subkeyword_of=kw1_eid, included_in=self.classif1_eid).eid
            kw3_eid = ce('Keyword', name=u'kw3', subkeyword_of=kw2_eid, included_in=self.classif1_eid).eid
            kw4_eid = ce('Keyword', name=u'kw4', subkeyword_of=kw3_eid, included_in=self.classif1_eid).eid
            ce('Keyword', name=u'kw5',  subkeyword_of=kw4_eid, included_in=self.classif1_eid).eid
            cnx.commit()
            cnx.execute('DELETE K subkeyword_of K3 WHERE K is Keyword, K eid %(kw3)s',
                        {'kw3': kw3_eid})
            cnx.commit()
            kw3 = cnx.find('Keyword', eid=kw3_eid).one()
            self.assertCountEqual([kw.name for kw in kw3.cw_adapt_to('ITree').iterparents()], [])
            self.assertCountEqual([kw.name for kw in kw3.descendant_of], [])
            self.assertCountEqual([kw.name for kw in kw3.reverse_descendant_of], ['kw5', 'kw4'])
            self.assertCountEqual([kw.name for kw in kw3.cw_adapt_to('ITree').recurse_children()], ['kw5', 'kw4'])

    def test_no_add_descendant_cycle(self):
        """no ``descendant_of`` cycle"""
        with self.admin_access.cnx() as cnx:
            ce = cnx.create_entity
            kw1 = ce('Keyword', name=u'kw1', included_in=self.classif1_eid)
            kw2 = ce('Keyword', name=u'kw2', subkeyword_of=kw1, included_in=self.classif1_eid)
            kw3 = ce('Keyword', name=u'kw3', subkeyword_of=kw2, included_in=self.classif1_eid)
            cnx.commit()
            rql = 'SET K1 descendant_of K3 WHERE K1 eid %(kw1)s, K3 eid %(kw3)s' % {'kw1': kw1.eid,  'kw3': kw3.eid}
            self.assertRaises(ValidationError, cnx.execute, rql)
            cnx.rollback()
            kw4 = ce('Keyword', name=u'kw4', included_in=self.classif1_eid)
            kw5 = ce('Keyword', name=u'kw4', subkeyword_of=kw4, included_in=self.classif1_eid)
            cnx.commit()
            with self.assertRaises(ValidationError):
                cnx.execute('SET K4 descendant_of K5 WHERE K4 eid %(kw4)s, K5 eid %(kw5)s',
                            {'kw4': kw4.eid,  'kw5': kw5.eid})


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



"""entity classes for classification schemes entities

:organization: Logilab
:copyright: 2007-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from logilab.common.deprecation import callable_deprecated

from cubicweb.entities import AnyEntity, fetch_config, adapters
from cubicweb.entity import EntityAdapter
from cubicweb.predicates import is_instance


class Classification(AnyEntity):
    __regid__ = 'Classification'
    fetch_attrs, cw_fetch_order = fetch_config(['name'])


class ClassificationITreeAdapter(EntityAdapter):
    __regid__ = 'ITree'
    __select__ = is_instance('Classification')

    def root(self):
        """returns the root object"""
        return None

    def parent(self):
        """returns the parent entity"""
        return None

    def iterparents(self):
        """returns parent entities"""
        yield self

    def children(self, entities=True):
        """returns the item's children"""
        return self.entity.related('included_in', 'object', entities=entities)

    def children_rql(self):
        """XXX returns RQL to get children"""
        return self.entity.cw_related_rql('included_in', 'object')

    def is_leaf(self):
        """returns true if this node as no child"""
        return bool(self.children())

    def is_root(self):
        """returns true if this node has no parent"""
        return True

    @callable_deprecated('[3.6] was specific to external project')
    def first_level_keywords(self):
        return self.req.execute('Any K,N ORDERBY N WHERE K included_in C, '
                                'NOT K subkeyword_of KK, K name N, '
                                'C eid %(x)s', {'x': self.eid})


class Keyword(AnyEntity):
    __regid__ = 'Keyword'
    fetch_attrs, cw_fetch_order = fetch_config(['name'])


class KeywordITreeAdapter(adapters.ITreeAdapter):
    __select__ = is_instance('Keyword', 'CodeKeyword')
    tree_relation = 'subkeyword_of'

    @property
    def classification(self):
        if self.entity.included_in:
            return self.entity.included_in[0]
        return None

    def parent(self):
        """ITree + IBreadcrumbs implementation"""
        try:
            return self.entity.related(self.tree_relation, self.child_role,
                                       entities=True)[0]
        except (KeyError, IndexError):
            return self.classification

    def iterparents(self):
        """returns parent keyword entities,
           without the root classification
        """
        if self.entity.subkeyword_of:
            parent = self.entity.subkeyword_of[0]
            while parent is not None:
                yield parent
                if parent.subkeyword_of:
                    parent = parent.subkeyword_of[0]
                else:
                    parent = None

    def iterchildren(self):
        """returns children entities"""
        if self.reverse_subkeyword_of:
            child = self.reverse_subkeyword_of[0]
            while child is not None:
                yield child
                if child.reverse_subkeyword_of:
                    child = child.reverse_subkeyword_of[0]
                else:
                    child = None

    def recurse_children(self, _done=None):
        """returns strict descendents"""
        if _done is not None and self.entity.eid in _done:
            return
        if _done is not None:
            _done.add(self.entity.eid)
            yield self.entity
        else:
            _done = set()
        for child in self.children():
            for entity in child.cw_adapt_to('ITree').recurse_children(_done):
                yield entity


class CodeKeyword(Keyword):
    __regid__ = 'CodeKeyword'
    rest_attr = 'code'
    fetch_attrs, cw_fetch_order = fetch_config(['code', 'name'])

    def dc_title(self):
        return u'%s - %s' % (self.code, self.name)



"""Notification views for keywords / classification schemes

:organization: Logilab
:copyright: 2003-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from logilab.common.decorators import cached

from cubicweb.predicates import is_instance
from cubicweb.sobjects.notification import NotificationView
from cubicweb import _


class KeywordNotificationView(NotificationView):
    __select__ = is_instance('Keyword')
    msgid_timestamp = True

    def recipients(self):
        """Returns the project's interested people (entities)"""
        creator = self.cw_rset.get_entity(0, 0).created_by[0]
        if not creator.is_in_group('managers') and creator.primary_email:
            return [(creator.primary_email[0].address, 'fr')]
        return []

    def context(self, **kwargs):
        context = NotificationView.context(self, **kwargs)
        entity = self.cw_rset.get_entity(0, 0)
        context['kw'] = entity.name
        return context


class KeywordNameChanged(KeywordNotificationView):
    __regid__ = 'notif_after_update_entity'

    content = _("keyword name changed from %(oldname)s to %(kw)s")

    @cached
    def get_oldname(self, entity):
        session = self.req
        try:
            return session.execute('Any N WHERE X eid %(x)s, X name N',
                                   {'x': entity.eid}, 'x')[0][0]
        except IndexError:
            return u'?'

    def context(self, **kwargs):
        entity = self.cw_rset.get_entity(0, 0)
        context = KeywordNotificationView.context(self, **kwargs)
        context['oldname'] = self.get_oldname(entity)
        return context

    def subject(self):
        entity = self.cw_rset.get_entity(0, 0)
        return self.req._('keyword name changed from %s to %s') % (
            self.get_oldname(entity), entity.name)



from logilab.common.testlib import unittest_main
from cubicweb.devtools.testlib import AutomaticWebTest


class AutomaticWebTest(AutomaticWebTest):
    no_auto_populate = ('Keyword', 'CodeKeyword')
    ignored_relations = set(('descendant_of',))

    def to_test_etypes(self):
        # only test the cube related entities
        return set(('Classification', 'Keyword', 'CodeKeyword'))

    def list_startup_views(self):
        return ()


if __name__ == '__main__':
    unittest_main()



# pylint: disable-msg=W0622
"""cubicweb-classification-schemes packaging information"""

modname = 'keyword'
distname = "cubicweb-keyword"

numversion = (2, 1, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
copyright = '''Copyright (c) 2003-2010 LOGILAB S.A. (Paris, FRANCE).
http://www.logilab.fr/ -- mailto:contact@logilab.fr'''

author = "Logilab"
author_email = "contact@logilab.fr"
web = 'https://forge.extranet.logilab.fr/cubicweb/cubes/%s' % distname

description = "classification schemes system for the Cubicweb framework"

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
    ]

__depends_cubes__ = {}
__depends__ = {'cubicweb': ">= 3.26.0, < 3.36.0"}



"""This cube handles classification schemes.

:organization: Logilab
:copyright: 2007-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
from yams.buildobjs import EntityType, RelationType, SubjectRelation, String

from cubicweb import _
from cubicweb.schema import RQLConstraint


class Classification(EntityType):
    """classification schemes are used by users to classify entities.
    """
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', ),
        'delete': ('managers', ),
        'update': ('managers', ),
    }
    name = String(required=True, fulltextindexed=True, unique=True, maxsize=128)
    classifies = SubjectRelation(
        'CWEType',
        # see relation type docstring
        constraints=[
            RQLConstraint(
                'RDEF to_entity O,'
                'RDEF relation_type R,'
                'R name "applied_to"',
                msg="Classification is trying to classifies an EntityType "
                "without applied_to relation"
            )
        ]
    )


class classifies(RelationType):
    """entity types classified by the classification. Only entity type
    supporting the applied_to relation can be selectioned
    """


class Keyword(EntityType):
    """A keyword is like a tag but is application specific
    and used to define a classification scheme
    """
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', 'users'),
        'delete': ('managers', ),
        'update': ('managers', ),
    }
    name = String(required=True, fulltextindexed=True, indexed=True, maxsize=128)

    subkeyword_of = SubjectRelation(
        'Keyword', cardinality='?*',
        description=_('which keyword (if any) this keyword specializes'),
        # if no included_in set, it'll be automatically added by a hook
        constraints=[RQLConstraint(
            'NOT S included_in CS1 OR EXISTS(S included_in CS2, O included_in CS2)'
        )])
    descendant_of = SubjectRelation('Keyword')
    included_in = SubjectRelation('Classification', cardinality='1*')


class CodeKeyword(Keyword):
    """A CodeKeyword is a keyword with a code and a name
    """
    __specializes_schema__ = True
    code = String(required=True, fulltextindexed=True, indexed=True, maxsize=128)


class subkeyword_of(RelationType):
    """a keyword can specialize another keyword"""


class included_in(RelationType):
    """a keyword is included in a classification scheme"""
    inlined = True


# define in parent application which entity types may be linked to a keyword
# by the applied_to relation

class applied_to(RelationType):
    """tagged objects"""
    fulltext_container = 'object'
    constraints = [RQLConstraint('S included_in CS, O is ET, CS classifies ET')]
